/*******************************************************************************
*									       *
* shell.c -- Nirvana Editor shell command execution			       *
*									       *
* Copyright (c) 1991-1997 Universities Research Association, Inc.	       *
* All rights reserved.							       *
* 									       *
* This material resulted from work developed under a Government Contract and   *
* is subject to the following license:  The Government retaFins a paid-up,     *
* nonexclusive, irrevocable worldwide license to reproduce, prepare derivative *
* works, perform publicly and display publicly by or for the Government,       *
* including the right to distribute to other Government contractors.  Neither  *
* the United States nor the United States Department of Energy, nor any of     *
* their employees, makes any warranty, express or implied, or assumes any      *
* legal liability or responsibility for the accuracy, completeness, or         *
* usefulness of any information, apparatus, product, or process disclosed, or  *
* represents that its use would not infringe privately owned rights.           *
*                                        				       *
* Fermilab Nirvana GUI Library						       *
* December, 1993							       *
*									       *
* Written by Mark Edel							       *
*									       *
*******************************************************************************/
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif
#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif
#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#ifdef _WIN32
# include <io.h>
# include <process.h>
# include <direct.h>
# include <windows.h>
# if 0
#  define USE_WIN32_HANDLES 1
# else
#  undef USE_WIN32_HANDLES
# endif
# if 1
#  define USE_WIN32_STDOUT_THREAD 1
# else
#  undef USE_WIN32_STDOUT_THREAD
# endif
# if 1
#  define USE_WIN32_STDERR_THREAD 1
# else
#  undef USE_WIN32_STDERR_THREAD
# endif
#endif
#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <Xm/PushBG.h>
#include <Xm/ScrolledW.h>
#include "../util/DialogF.h"
#include "../util/misc.h"
#include "textBuf.h"
#include "text.h"
#include "nedit.h"
#include "window.h"
#include "preferences.h"
#include "file.h"
#include "shell.h"
#include "macro.h"
#include "interpret.h"

/* Tuning parameters */
#define IO_BUF_SIZE 4096	/* size of buffers for collecting cmd output */
#define MAX_SHELL_CMD_LEN 2048	/* max length of a shell command (should be
				   eliminated, but substitutePercent needs) */
#define MAX_OUT_DIALOG_ROWS 30	/* max height of dialog for command output */
#define MAX_OUT_DIALOG_COLS 80	/* max width of dialog for command output */
#define OUTPUT_FLUSH_FREQ 1000	/* how often (msec) to flush output buffers
    	    	    	    	   when process is taking too long */
#define BANNER_WAIT_TIME 6000	/* how long to wait (msec) before putting up
    	    	    	    	   Shell Command Executing... banner */

/* flags for issueCommand */
#define ACCUMULATE 1
#define ERROR_DIALOGS 2
#define REPLACE_SELECTION 4
#define RELOAD_FILE_AFTER 8
#define OUTPUT_TO_DIALOG 16
#define OUTPUT_TO_STRING 32
#define OUTPUT_TO_NEW_WINDOW 64

/* element of a buffer list for collecting output from shell processes */
typedef struct bufElem {
    struct bufElem *next;
    int length;
    char contents[IO_BUF_SIZE];
} buffer;

/* data attached to window during shell command execution with
   information for controling and communicating with the process */
typedef struct {
    int flags;
#ifdef USE_WIN32_HANDLES
    HANDLE stdinFD, stdoutFD, stderrFD;
#else
    int stdinFD, stdoutFD, stderrFD;
#endif
    pid_t childPid;
    buffer *outBufs, *errBufs;
    char *input;
    char *inPtr;
    Widget textW;
    int leftPos, rightPos;
    int inLength;
    XtIntervalId bannerTimeoutID, flushTimeoutID;
    char bannerIsUp;
    char fromMacro;
#ifndef _WIN32
	XtInputId stdinInputID;
	XtInputId stdoutInputID;
	XtInputId stderrInputID;
#else
	HANDLE stdinThreadHandle;
	HANDLE stdinThreadExitEventHandle;
	HANDLE stdoutThreadHandle;
	HANDLE stdoutThreadExitEventHandle;
	HANDLE stderrThreadHandle;
	HANDLE stderrThreadExitEventHandle;
	BOOLEAN terminatedOnError;
#endif
} shellCmdInfo;

static void issueCommand(WindowInfo *window, char *command, char *input,
	int inputLen, int flags, Widget textW, int replaceLeft,
	int replaceRight, int fromMacro);
static void stdoutReadProc(XtPointer clientData, int *source, XtInputId *id);
static void stderrReadProc(XtPointer clientData, int *source, XtInputId *id);
static void stdinWriteProc(XtPointer clientData, int *source, XtInputId *id);
static void finishCmdExecution(WindowInfo *window, int terminatedOnError);
static pid_t forkCommand(Widget parent, char *command, char *cmdDir,
#ifdef USE_WIN32_HANDLES
	HANDLE *stdinFD, HANDLE *stdoutFD, HANDLE *stderrFD);
#else
	int *stdinFD, int *stdoutFD, int *stderrFD);
#endif
static void addOutput(buffer **bufList, buffer *buf);
static char *coalesceOutput(buffer **bufList, int *length);
static void freeBufList(buffer **bufList);
static void removeTrailingNewlines(char *string);
static Widget createOutputDialog(WindowInfo *window, char *text, 
	unsigned char dialogStyle);
static void destroyOutDialogCB(Widget w, XtPointer callback, XtPointer closure);
static void measureText(char *text, int wrapWidth, int *rows, int *cols);
static void truncateString(char *string, int length);
static int substitutePercent(WindowInfo *window, char *outStr, char *inStr,
        int outLen, char *selection);
static void bannerTimeoutProc(XtPointer clientData, XtIntervalId *id);
static void flushTimeoutProc(XtPointer clientData, XtIntervalId *id);

/*
** Filter the current selection through shell command "command".  The selection
** is removed, and replaced by the output from the command execution.  Failed
** command status and output to stderr are presented in dialog form.
*/
void FilterSelection(WindowInfo *window, char *command, int fromMacro)
{
    int left, right, textLen;
    char *text;
    char subsCommand[MAX_SHELL_CMD_LEN];

    /* Can't do two shell commands at once in the same window */
    if (window->shellCmdData != NULL) {
    	XBell(TheDisplay, 0);
    	return;
    }

    /* Get the selection and the range in character positions that it
       occupies.  Beep and return if no selection */
    text = BufGetSelectionText(window->editorInfo->buffer);
    if (*text == '\0') {
	XtFree(text);
	XBell(TheDisplay, 0);
	return;
    }
    
/* .b: Dialog message is now inside function */
    if (!substitutePercent(window, subsCommand, command, 
		MAX_SHELL_CMD_LEN, text))
        return;
/* .e */
    	
    textLen = strlen(text);
    BufUnsubstituteNullChars(text, window->editorInfo->buffer);
    left = window->editorInfo->buffer->primary.start;
    right = window->editorInfo->buffer->primary.end;
    
    /* Issue the command and collect its output */
    issueCommand(window, subsCommand, text, textLen, ACCUMULATE | ERROR_DIALOGS |
	    REPLACE_SELECTION, window->lastFocus, left, right, fromMacro);
}

/*
** Execute shell command "command", depositing the result at the current
** insert position or in the current selection if if the window has a
** selection.
*/
void ExecShellCommand(WindowInfo *window, char *command, int fromMacro)
{
    int left, right, flags = 0;
    char subsCommand[MAX_SHELL_CMD_LEN];
	
    /* Can't do two shell commands at once in the same window */
    if (window->shellCmdData != NULL) {
    	XBell(TheDisplay, 0);
    	return;
    }
    
    /* get the selection or the insert position */
    if (GetSimpleSelection(window->editorInfo->buffer, &left, &right))
    	flags = ACCUMULATE | REPLACE_SELECTION;
    else
    	left = right = TextGetCursorPos(window->lastFocus);
    
    if (!substitutePercent(window, subsCommand, command,
            MAX_SHELL_CMD_LEN, NULL))
        return;
    	
    /* issue the command */
    issueCommand(window, subsCommand, NULL, 0, flags, window->lastFocus, left,
	    right, fromMacro);
}

/*
** Execute shell command "command", on input string "input", depositing the
** in a macro string (via a call back to ReturnShellCommandOutput).
*/
void ShellCmdToMacroString(WindowInfo *window, char *command, char *input)
{
    char *inputCopy;
    char subsCommand[MAX_SHELL_CMD_LEN];
    
    /* Make a copy of the input string for issueCommand to hold and free
       upon completion */
    inputCopy = *input == '\0' ? NULL : XtNewString(input);
    
    if (!substitutePercent(window, subsCommand, command,
            MAX_SHELL_CMD_LEN, inputCopy))
        return;

    /* fork the command and begin processing input/output */
    issueCommand(window, subsCommand, inputCopy, strlen(input),
	    ACCUMULATE | OUTPUT_TO_STRING, NULL, 0, 0, True);
}

/*
** Execute the line of text where the the insertion cursor is positioned
** as a shell command.
*/
void ExecCursorLine(WindowInfo *window, int fromMacro)
{
    char *cmdText;
    int left, right, insertPos;

    /* Can't do two shell commands at once in the same window */
    if (window->shellCmdData != NULL) {
    	XBell(TheDisplay, 0);
    	return;
    }

    /* get all of the text on the line with the insert position */
    if (!GetSimpleSelection(window->editorInfo->buffer, &left, &right)) {
	left = right = TextGetCursorPos(window->lastFocus);
	left = BufStartOfLine(window->editorInfo->buffer, left);
	right = BufEndOfLine(window->editorInfo->buffer, right);
	insertPos = right;
    } else
    	insertPos = BufEndOfLine(window->editorInfo->buffer, right);
    cmdText = BufGetRange(window->editorInfo->buffer, left, right);
    BufUnsubstituteNullChars(cmdText, window->editorInfo->buffer);
    
    /* insert a newline after the entire line */
    BufInsert(window->editorInfo->buffer, insertPos, "\n");

    /* issue the command */
    issueCommand(window, cmdText, NULL, 0, 0, window->lastFocus, insertPos+1,
	    insertPos+1, fromMacro);

    XtFree(cmdText);
}

/*
** Do a shell command, with the options allowed to users (input source,
** output destination, save first and load after) in the shell commands
** menu.
*/
void DoShellMenuCmd(WindowInfo *window, char *command, int input, int output,
	int outputReplacesInput, int saveFirst, int loadAfter, int fromMacro) 
{
    int flags = 0;
    char *text;
    char subsCommand[MAX_SHELL_CMD_LEN];
    int left, right, textLen;
    WindowInfo *inWindow = window;
    Widget outWidget;

    /* Can't do two shell commands at once in the same window */
    if (window->shellCmdData != NULL) {
    	XBell(TheDisplay, 0);
    	return;
    }

    /* Get the command input as a text string.  If there is input, errors
      shouldn't be mixed in with output, so set flags to ERROR_DIALOGS */
    if (input == FROM_SELECTION) {
	text = BufGetSelectionText(window->editorInfo->buffer);
	if (*text == '\0') {
    	    XtFree(text);
    	    XBell(TheDisplay, 0);
    	    return;
    	}
    	flags |= ACCUMULATE | ERROR_DIALOGS;
    } else if (input == FROM_WINDOW) {
	text = BufGetAll(window->editorInfo->buffer);
    	flags |= ACCUMULATE | ERROR_DIALOGS;
    } else if (input == FROM_EITHER) {
	text = BufGetSelectionText(window->editorInfo->buffer);
	if (*text == '\0') {
	    XtFree(text);
	    text = BufGetAll(window->editorInfo->buffer);
    	}
    	flags |= ACCUMULATE | ERROR_DIALOGS;
    } else /* FROM_NONE */
    	text = NULL;
    
    /* If the buffer was substituting another character for ascii-nuls,
       put the nuls back in before exporting the text */
    if (text != NULL) {
    	textLen = strlen(text);
    	BufUnsubstituteNullChars(text, window->editorInfo->buffer);
    } else
    	textLen = 0;
    
    /* Assign the output destination.  If output is to a new window,
       create it, and run the command from it instead of the current
       one, to free the current one from waiting for lengthy execution */
    if (output == TO_DIALOG) {
    	outWidget = NULL;
    	flags |= OUTPUT_TO_DIALOG;
    	left = right = 0;
    } else if (output == TO_NEW_WINDOW) {
    	inWindow = EditNewFile(window);
    	outWidget = inWindow->textArea;
    	flags |= OUTPUT_TO_NEW_WINDOW;
    	left = right = 0;
    } else { /* TO_SAME_WINDOW */
    	outWidget = window->lastFocus;
    	if (outputReplacesInput && input != FROM_NONE) {
    	    if (input == FROM_WINDOW) {
    		left = 0;
    		right = window->editorInfo->buffer->length;
    	    } else if (input == FROM_SELECTION) {
    	    	GetSimpleSelection(window->editorInfo->buffer, &left, &right);
	        flags |= ACCUMULATE | REPLACE_SELECTION;
    	    } else if (input == FROM_EITHER) {
    	    	if (GetSimpleSelection(window->editorInfo->buffer, &left, &right))
	            flags |= ACCUMULATE | REPLACE_SELECTION;
	        else {
	            left = 0;
	            right = window->editorInfo->buffer->length;
	        }
	    }
    	} else {
	    if (GetSimpleSelection(window->editorInfo->buffer, &left, &right))
	        flags |= ACCUMULATE | REPLACE_SELECTION;
	    else
    		left = right = TextGetCursorPos(window->lastFocus);
    	}
    }
    
    /* Catch errors in the command if the user wants the file loaded
    ** after the command. */
    if (loadAfter) {
    	flags |= ACCUMULATE | ERROR_DIALOGS;
    }
    
    /* If the command requires the file be saved first, save it */
    if (saveFirst) {
    	if (!SaveWindow(window)) {
    	    if (input != FROM_NONE)
    		XtFree(text);
    	    return;
	}
    }
    
/* .b: Dialog message is now inside function */
    if (!substitutePercent(window, subsCommand, command,
            MAX_SHELL_CMD_LEN, text))
        return;
/* .e */
    	
    /* If the command requires the file to be reloaded after execution, set
       a flag for issueCommand to deal with it when execution is complete */
    if (loadAfter)
    	flags |= RELOAD_FILE_AFTER;
    	
    /* issue the command */
    issueCommand(inWindow, subsCommand, text, textLen, flags, outWidget, left,
	    right, fromMacro);
}

/*
** Cancel the shell command in progress
*/
void AbortShellCommand(WindowInfo *window)
{
    shellCmdInfo *cmdData = window->shellCmdData;

    if (cmdData == NULL)
    	return;
#ifdef _WIN32
	TerminateProcess((HANDLE)cmdData->childPid, 0);
#else
	kill(- cmdData->childPid, SIGTERM);
#endif
#ifdef _WIN32
	// signal the threads to finish.
	if(cmdData->stdinThreadExitEventHandle) {
		SetEvent(cmdData->stdinThreadExitEventHandle);
	}
	if(cmdData->stdoutThreadExitEventHandle) {
		SetEvent(cmdData->stdoutThreadExitEventHandle);
	}
	if(cmdData->stderrThreadExitEventHandle) {
		SetEvent(cmdData->stderrThreadExitEventHandle);
	}
#else
    finishCmdExecution(window, True);
#endif
}

/*
** Function to check if there is a shell command active for window.
*/
Boolean IsShellCommandInProgress(WindowInfo *window)
{
	return (window->shellCmdData != NULL);
}

#ifdef _WIN32
/* Thread to write stdin. If it blocks who cares. */
static DWORD WINAPI stdinWriteThread(LPVOID lpParam) {
    WindowInfo *window = (WindowInfo *)lpParam;
    shellCmdInfo *cmdData = window->shellCmdData;

	// No file descriptor to write on so just return;
#ifndef USE_WIN32_HANDLES
	if(cmdData->stdinFD == -1) {
		return 0;
	}
#else
#endif

	while(1) {
		int nWritten;

		/* We have been signaled to terminate */
		if(WaitForSingleObject(cmdData->stdinThreadExitEventHandle, 0/*milliseconds*/) 
			== WAIT_OBJECT_0) {
			cmdData->terminatedOnError = TRUE;
			break;
		}
	

#ifdef USE_WIN32_HANDLES
		if(!WriteFile(cmdData->stdinFD, cmdData->inPtr, cmdData->inLength, &nWritten, NULL)) {
#else
		nWritten = write(cmdData->stdinFD, cmdData->inPtr, cmdData->inLength);
		if (nWritten == -1) {
#endif
#if 0
			PERROR("NEdit: Write to shell command failed");
#endif
			cmdData->terminatedOnError = TRUE;
			break;
		} else {
			cmdData->inPtr += nWritten;
			cmdData->inLength -= nWritten;
			if (cmdData->inLength <= 0) {
#ifdef USE_WIN32_HANDLES
				CloseHandle(cmdData->stdinFD);
#else
    			if(cmdData->stdinFD != -1) {
					close(cmdData->stdinFD);
					cmdData->stdinFD = -1;
				}
#endif
    			cmdData->inPtr = NULL;
				break;
    		}
		}
	}
	return 0;
}

/* Thread to read stdout. If it blocks who cares. */
static DWORD WINAPI stdoutReadThread(LPVOID lpParam) {
    WindowInfo *window = (WindowInfo *)lpParam;
    shellCmdInfo *cmdData = window->shellCmdData;
 
	while(1) {
		buffer *buf;
		int nRead;
    
		/* See if we have been signaled to terminate */
		if(WaitForSingleObject(cmdData->stdoutThreadExitEventHandle, 0/*milliseconds*/) 
			== WAIT_OBJECT_0) {
			cmdData->terminatedOnError = TRUE;
			break;
		}

		buf = (buffer *)XtMalloc(sizeof(buffer));

		/* read from the process' stdout stream */
#ifdef USE_WIN32_HANDLES
	    if(!ReadFile(cmdData->stdoutFD, buf->contents, IO_BUF_SIZE, &nRead, NULL)) {
#else
		nRead = read(cmdData->stdoutFD, buf->contents, IO_BUF_SIZE);
		if (nRead == -1) { /* error */
#endif
#if 0
			PERROR("NEdit: Error reading shell command output stream");
#endif
			cmdData->terminatedOnError = TRUE;
			XtFree((char *)buf);
			break;
		}
    
		/* end of data.  If the stdout stream is done too, execution of the
		   shell process is complete, and we can display the results */
		if (nRead == 0) {
			XtFree((char *)buf);
	    	break;
	    }
    
		/* characters were read successfully, add buf to linked list of buffers */
		buf->length = nRead;
		addOutput(&cmdData->outBufs, buf);
	}
	
	return 0;
}

/* Thread to read stdout. If it blocks who cares. */
static DWORD WINAPI stderrReadThread(LPVOID lpParam) {
    WindowInfo *window = (WindowInfo *)lpParam;
    shellCmdInfo *cmdData = window->shellCmdData;

	while(1) {
	    buffer *buf;
	    int nRead;
    
		/* See if we have been signaled to terminate */
		if(WaitForSingleObject(cmdData->stderrThreadExitEventHandle, 0/*milliseconds*/) 
			== WAIT_OBJECT_0) {
			cmdData->terminatedOnError = TRUE;
			break;
		}

		buf = (buffer *)XtMalloc(sizeof(buffer));

		/* read from the process' stderr stream */
#ifdef USE_WIN32_HANDLES
	    if(!ReadFile(cmdData->stderrFD, buf->contents, IO_BUF_SIZE, &nRead, NULL)) {
#else
		nRead = read(cmdData->stderrFD, buf->contents, IO_BUF_SIZE);
		if (nRead == -1) { /* error */
#endif
#if 0
			PERROR("NEdit: Error reading shell command error stream");
#endif
			cmdData->terminatedOnError = TRUE;
			XtFree((char *)buf);
			break;
		}
    
		/* end of data.  If the stdout stream is done too, execution of the
		   shell process is complete, and we can display the results */
		if (nRead == 0) {
			XtFree((char *)buf);
	    	break;
	    }
    
		/* characters were read successfully, add buf to linked list of buffers */
		buf->length = nRead;
		addOutput(&cmdData->errBufs, buf);
	}
	
	return 0;
}

// Thread to handle command completion.
// Waits for stdin, stderr, and stdout threads to complete.
static DWORD WINAPI commandCompletionThread(LPVOID lpParam) {
    WindowInfo *window = (WindowInfo *)lpParam;
    shellCmdInfo *cmdData = window->shellCmdData;
	HANDLE handlelist[3];
	INT n;
	if(cmdData->stdinThreadHandle) {
		handlelist[n] = cmdData->stdinThreadHandle;
		n++;
	}
	if(cmdData->stdoutThreadExitEventHandle) {
		handlelist[n] = cmdData->stdoutThreadHandle;
		n++;
	}
	if(cmdData->stderrThreadExitEventHandle) {
		handlelist[n] = cmdData->stderrThreadHandle;
		n++;
	}
	WaitForMultipleObjects(n, handlelist, TRUE, INFINITE);
    finishCmdExecution(window, False);
	return 0;
}

static void createStdinWriteThread(WindowInfo *window) {
	unsigned dwThreadId;
	shellCmdInfo *cmdData = window->shellCmdData;

	cmdData->stdinThreadHandle = (HANDLE)_beginthreadex(
		NULL,                        // no security attributes 
	    0,                           // use default stack size  
		stdinWriteThread,            // thread function 
		window,                      // argument to thread function 
		CREATE_SUSPENDED,            // The thread is suspended until we start the
			                         // child process
		&dwThreadId                  // returns the thread identifier
	);
	if(cmdData->stdinThreadHandle == 0) {
		PERROR("NEdit: Error creating stdin write thread.");
	}

	// Create an event to use to signal the thread to terminate.
	cmdData->stdinThreadExitEventHandle = CreateEvent(
		NULL,                    // pointer to security attributes 
		FALSE,                   // flag for manual-reset event 
		FALSE,                   // flag for initial state 
		NULL                     // pointer to event-object name (unnamed)
	);
}

static void createStdoutReadThread(WindowInfo *window) {
	DWORD dwThreadId;
	shellCmdInfo *cmdData = window->shellCmdData;

	cmdData->stdoutThreadHandle = (HANDLE)_beginthreadex(
		NULL,                        // no security attributes 
	    0,                           // use default stack size  
		stdoutReadThread,            // thread function 
		window,                      // argument to thread function 
		CREATE_SUSPENDED,            // The thread is suspended until we start the
			                         // child process
		&dwThreadId                  // returns the thread identifier
	);
	if(cmdData->stdoutThreadHandle == 0) {
		PERROR("NEdit: Error creating stdout read thread.");
	}

	// Create an event to use to signal the thread to terminate.
	cmdData->stdoutThreadExitEventHandle = CreateEvent(
		NULL,                    // pointer to security attributes 
		FALSE,                   // flag for manual-reset event 
		FALSE,                   // flag for initial state 
		NULL                     // pointer to event-object name (unnamed)
	);
}

static void createStderrReadThread(WindowInfo *window) {
	unsigned dwThreadId;
	shellCmdInfo *cmdData = window->shellCmdData;

	cmdData->stderrThreadHandle = (HANDLE)_beginthreadex(
		NULL,                        // no security attributes 
	    0,                           // use default stack size  
		stderrReadThread,            // thread function 
		window,                      // argument to thread function 
		CREATE_SUSPENDED,            // The thread is suspended until we start the
			                         // child process
		&dwThreadId                  // returns the thread identifier
	);
	if(cmdData->stderrThreadHandle == 0) {
		PERROR("NEdit: Error creating stderr read thread.");
	}
		
	// Create an event to use to signal the thread to terminate.
	cmdData->stderrThreadExitEventHandle = CreateEvent(
		NULL,                    // pointer to security attributes 
		FALSE,                   // flag for manual-reset event 
		FALSE,                   // flag for initial state 
		NULL                     // pointer to event-object name (unnamed)
	);
}
#endif

/*
** Issue a shell command and feed it the string "input".  Output can be
** directed either to text widget "textW" where it replaces the text between
** the positions "replaceLeft" and "replaceRight", to a separate pop-up dialog
** (OUTPUT_TO_DIALOG), or to a macro-language string (OUTPUT_TO_STRING).  If
** "input" is NULL, no input is fed to the process.  If an input string is
** provided, it is freed when the command completes.  Flags:
**
**   ACCUMULATE     	Causes output from the command to be saved up until
**  	    	    	the command completes.
**   ERROR_DIALOGS  	Presents stderr output separately in popup a dialog,
**  	    	    	and also reports failed exit status as a popup dialog
**  	    	    	including the command output.
**   REPLACE_SELECTION  Causes output to replace the selection in textW.
**   RELOAD_FILE_AFTER  Causes the file to be completely reloaded after the
**  	    	    	command completes.
**   OUTPUT_TO_DIALOG   Send output to a pop-up dialog instead of textW
**   OUTPUT_TO_NEW_WINDOW   textW is in a new window
**   OUTPUT_TO_STRING   Output to a macro-language string instead of a text
**  	    	    	widget or dialog.
**
** REPLACE_SELECTION, ERROR_DIALOGS, and OUTPUT_TO_STRING can only be used
** along with ACCUMULATE (these operations can't be done incrementally).
*/
static void issueCommand(WindowInfo *window, char *command, char *input,
	int inputLen, int flags, Widget textW, int replaceLeft,
	int replaceRight, int fromMacro)
{
#ifdef USE_WIN32_HANDLES
    HANDLE stdinFD, stdoutFD, stderrFD;
#else
    int stdinFD = -1, stdoutFD = -1, stderrFD = -1;
#endif
    XtAppContext context = XtWidgetToApplicationContext(window->shell);
    shellCmdInfo *cmdData;
    pid_t childPid;
    
    /* verify consistency of input parameters */
    if ((flags & ERROR_DIALOGS || flags & REPLACE_SELECTION ||
	    flags & OUTPUT_TO_STRING) && !(flags & ACCUMULATE))
    	return;
    
    /* a shell command called from a macro must be executed in the same
       window as the macro, regardless of where the output is directed,
       so the user can cancel them as a unit */
    if (fromMacro)
    	window = MacroRunWindow();
    
    /* put up a watch cursor over the waiting window */
    if (!fromMacro)
    	BeginWait(window->shell);
    
    /* enable the cancel menu item */
    if (!fromMacro)
    	XtSetSensitive(window->cancelShellItem, True);

    /* fork the subprocess and issue the command */
    childPid = forkCommand(window->shell, command, window->editorInfo->path, &stdinFD,
	    &stdoutFD, (flags & ERROR_DIALOGS) ? &stderrFD : NULL);
    
	if(childPid == -1) {
	    if (!fromMacro)
			XtSetSensitive(window->cancelShellItem, False);
		EndWait(window->shell);
		return;
	}

#ifndef _WIN32
    /* set the pipes connected to the process for non-blocking i/o */
	if (fcntl(stdinFD, F_SETFL, O_NONBLOCK) < 0) {
    	PERROR("NEdit: Internal error (fcntl)");
		return -1;
	}
    if (fcntl(stdoutFD, F_SETFL, O_NONBLOCK) < 0) {
    	PERROR("NEdit: Internal error (fcntl1)");
		return -1;
	}
    if (flags & ERROR_DIALOGS) {
		if (fcntl(stderrFD, F_SETFL, O_NONBLOCK) < 0) {
    	    PERROR("NEdit: Internal error (fcntl2)");
			return -1;
		}
    }
#endif
    
    /* if there's nothing to write to the process' stdin, close it now */
    if (input == NULL) {
		if(stdinFD != -1) {
			close(stdinFD);
			stdinFD = -1;
		}
    }
	
    /* Create a data structure for passing process information around
       amongst the callback routines which will process i/o and completion */
    cmdData = (shellCmdInfo *)XtMalloc(sizeof(shellCmdInfo));
    window->shellCmdData = cmdData;
    cmdData->flags = flags;
    cmdData->stdinFD = stdinFD;
    cmdData->stdoutFD = stdoutFD;
    cmdData->stderrFD = stderrFD;
    cmdData->childPid = childPid;
    cmdData->outBufs = NULL;
    cmdData->errBufs = NULL;
    cmdData->input = input;
    cmdData->inPtr = input;
    cmdData->textW = textW;
    cmdData->bannerIsUp = False;
    cmdData->fromMacro = fromMacro;
    cmdData->leftPos = replaceLeft;
    cmdData->rightPos = replaceRight;
    cmdData->inLength = inputLen;
    
    /* Set up timer proc for putting up banner when process takes too long */
    if (fromMacro)
    	cmdData->bannerTimeoutID = 0;
    else
    	cmdData->bannerTimeoutID = XtAppAddTimeOut(context, BANNER_WAIT_TIME,
    	    	bannerTimeoutProc, window);

    /* Set up timer proc for flushing output buffers periodically */
	if ((flags & ACCUMULATE) || textW == NULL) {
    	cmdData->flushTimeoutID = 0;
	}
    else {
		cmdData->flushTimeoutID = XtAppAddTimeOut(context, OUTPUT_FLUSH_FREQ,
	    	flushTimeoutProc, window);
	}
    	
#ifndef _WIN32
    /* set up callbacks for activity on the file descriptors */
	cmdData->stdoutInputID = XtAppAddInput(context, stdoutFD, (XtPointer)
			XtInputReadMask,
			stdoutReadProc, window);

    if (input != NULL) {
    	cmdData->stdinInputID = XtAppAddInput(context, stdinFD,
    	    	(XtPointer)XtInputWriteMask, stdinWriteProc, window);
	}
    else
    	cmdData->stdinInputID = 0;

    if (flags & ERROR_DIALOGS) {
		cmdData->stderrInputID = XtAppAddInput(context, stderrFD, (XtPointer)
			XtInputReadMask,
    		stderrReadProc, window);
    }
    else
    	cmdData->stderrInputID = 0;
    
    /* If this was called from a macro, preempt the macro until shell
       command completes */
    if (fromMacro)
    	PreemptMacro();

#else
	cmdData->terminatedOnError = FALSE;
	cmdData->stdinThreadHandle = 0;
	cmdData->stdinThreadExitEventHandle = 0;
	cmdData->stdoutThreadHandle = 0;
	cmdData->stdoutThreadExitEventHandle = 0;
	cmdData->stderrThreadHandle = 0;
	cmdData->stderrThreadExitEventHandle = 0;

	createStdoutReadThread(window);
    if (input != NULL) {
		createStdinWriteThread(window);
	}
    if (flags & ERROR_DIALOGS) {
		createStderrReadThread(window);
	}
//	createCommandCompletionThread(window);
	// Start the threads now that we are all setup.
	ResumeThread(cmdData->stdinThreadHandle);
	ResumeThread(cmdData->stdoutThreadHandle);
	ResumeThread(cmdData->stderrThreadHandle);

	// Wait here for the command and threads to finish;
	{
	HANDLE handlelist[4];
	INT n = 0;
	if(cmdData->stdinThreadHandle) {
		handlelist[n] = cmdData->stdinThreadHandle;
		n++;
	}
	if(cmdData->stdoutThreadHandle) {
		handlelist[n] = cmdData->stdoutThreadHandle;
		n++;
	}
	if(cmdData->stderrThreadHandle) {
		handlelist[n] = cmdData->stderrThreadHandle;
		n++;
	}
	if(cmdData->childPid) {
		handlelist[n] = cmdData->childPid;
		n++;
	}
	while(1) {
		DWORD rc;
#if 0
		rc = MsgWaitForMultipleObjects(n, handlelist, TRUE, 0, QS_ALLINPUT);
		if(rc == WAIT_TIMEOUT) {
			// keep looping.
		} else if((rc >= WAIT_OBJECT_0 && rc < (WAIT_OBJECT_0 + n)) 
			   || (rc >= WAIT_ABANDONED_0  && rc < (WAIT_OBJECT_0 + n))
		) {
			break;
		}
#else
		rc = WaitForMultipleObjects(n, handlelist, TRUE, 0);
		if(rc != WAIT_TIMEOUT) {
			break;
		}
#endif
    	XtAppProcessEvent(XtWidgetToApplicationContext(window->shell), XtIMAll);
	}
	}
	finishCmdExecution(window, cmdData->terminatedOnError);
#endif
}

#ifndef _WIN32
/*
** Called when the shell sub-process stdout stream has data.  Reads data into
** the "outBufs" buffer chain in the window->shellCommandData data structure.
*/
static void stdoutReadProc(XtPointer clientData, int *source, XtInputId *id)
{
    WindowInfo *window = (WindowInfo *)clientData;
    shellCmdInfo *cmdData = window->shellCmdData;
    buffer *buf;
    int nRead;

    /* read from the process' stdout stream */
    buf = (buffer *)XtMalloc(sizeof(buffer));
    
	nRead = read(cmdData->stdoutFD, buf->contents, IO_BUF_SIZE);
    if (nRead == -1) { /* error */
		if (errno != EWOULDBLOCK && errno != EAGAIN) {
			PERROR("NEdit: Error reading shell command output");
			XtFree((char *)buf);
			finishCmdExecution(window, True);
		}
		return;
    }
    
    /* end of data.  If the stderr stream is done too, execution of the
       shell process is complete, and we can display the results */
    if (nRead == 0) {
    	XtFree((char *)buf);
		if(cmdData->stdoutInputID) {
    		XtRemoveInput(cmdData->stdoutInputID);
    		cmdData->stdoutInputID = 0;
		}
    	if (cmdData->stderrInputID == 0)
    	    finishCmdExecution(window, False);
    	return;
    }
    
    /* characters were read successfully, add buf to linked list of buffers */
    buf->length = nRead;
    addOutput(&cmdData->outBufs, buf);
}

/*
** Called when the shell sub-process stderr stream has data.  Reads data into
** the "errBufs" buffer chain in the window->shellCommandData data structure.
*/
static void stderrReadProc(XtPointer clientData, int *source, XtInputId *id)
{
    WindowInfo *window = (WindowInfo *)clientData;
    shellCmdInfo *cmdData = window->shellCmdData;
    buffer *buf;
    int nRead;
    
    /* read from the process' stderr stream */
    buf = (buffer *)XtMalloc(sizeof(buffer));
    nRead = read(cmdData->stderrFD, buf->contents, IO_BUF_SIZE);
    if (nRead == -1) { /* error */
		if (errno != EWOULDBLOCK && errno != EAGAIN) {
			PERROR("NEdit: Error reading shell command error stream");
		    XtFree((char *)buf);
			finishCmdExecution(window, True);
		}
		return;
    }
    
    /* end of data.  If the stdout stream is done too, execution of the
       shell process is complete, and we can display the results */
    if (nRead == 0) {
    	XtFree((char *)buf);
		if(cmdData->stderrInputID) {
    		XtRemoveInput(cmdData->stderrInputID);
    		cmdData->stderrInputID = 0;
		}
    	if (cmdData->stdoutInputID == 0)
    	    finishCmdExecution(window, False);
    	return;
    }
    
    /* characters were read successfully, add buf to linked list of buffers */
    buf->length = nRead;
    addOutput(&cmdData->errBufs, buf);
}
#endif

/*
** Called when the shell sub-process stdin stream is ready for input.  Writes
** data from the "input" text string passed to issueCommand.
*/
static void stdinWriteProc(XtPointer clientData, int *source, XtInputId *id)
{
    WindowInfo *window = (WindowInfo *)clientData;
    shellCmdInfo *cmdData = window->shellCmdData;
    int nWritten;

#ifdef USE_WIN32_HANDLES
    if(!WriteFile(cmdData->stdinFD, cmdData->inPtr, cmdData->inLength, &nWritten, NULL)) {
		if(1) {
#else
	if(cmdData->stdinFD == -1) {
		return;
	}

    nWritten = write(cmdData->stdinFD, cmdData->inPtr, cmdData->inLength);
    if (nWritten == -1) {
# ifdef _WIN32
		if (errno != EAGAIN) {
# else
    	if (errno != EWOULDBLOCK && errno != EAGAIN) {
# endif
#endif
    	    PERROR("NEdit: Write to shell command failed");
    	    finishCmdExecution(window, True);
    	}
    } else {
	cmdData->inPtr += nWritten;
	cmdData->inLength -= nWritten;
	if (cmdData->inLength <= 0) {
#ifdef USE_WIN32_HANDLES
			CloseHandle(cmdData->stdinFD);
#else
    	    if(cmdData->stdinFD != -1) {
				close(cmdData->stdinFD);
				cmdData->stdinFD = -1;
			}
#endif
    	    cmdData->inPtr = NULL;
#ifdef _WIN32
			// signal the thread to exit.
			if(cmdData->stdinThreadExitEventHandle) {
				SetEvent(cmdData->stdinThreadExitEventHandle);
			}
#else
			if(cmdData->stdinInputID) {
				XtRemoveInput(cmdData->stdinInputID);
				cmdData->stdinInputID = 0;
			}
#endif
    	}
    }
}

/*
** Timer proc for putting up the "Shell Command in Progress" banner if
** the process is taking too long.
*/
static void bannerTimeoutProc(XtPointer clientData, XtIntervalId *id)
{
    WindowInfo *window = (WindowInfo *)clientData;
    shellCmdInfo *cmdData = window->shellCmdData;
    
    cmdData->bannerIsUp = True;
    SetModeMessage(window,
    	    "Shell Command in Progress -- Press Ctrl+. to Cancel");
    cmdData->bannerTimeoutID = 0;
}

/*
** Timer proc for flushing output buffers periodically when the process
** takes too long.
*/
static void flushTimeoutProc(XtPointer clientData, XtIntervalId *id)
{
    WindowInfo *window = (WindowInfo *)clientData;
    shellCmdInfo *cmdData = window->shellCmdData;
	textBuffer *buf;
    int len;
    char *outText;
    
	if(cmdData == NULL) {
#if 0
		DialogF(DF_WARN, window->shell, 1, "flushTimeoutProc(): cmdData == 0", "OK");
#endif
		return;
	}
    
    buf = TextGetBuffer(cmdData->textW);

    /* shouldn't happen, but it would be bad if it did */
    if (cmdData->textW == NULL)
    	return;

	outText = coalesceOutput(&cmdData->outBufs, &len);
	if (len != 0) {
		if (BufSubstituteNullChars(outText, len, buf)) {
			BufReplace(buf, cmdData->leftPos, cmdData->rightPos, outText);
			TextSetCursorPos(cmdData->textW, cmdData->leftPos+strlen(outText));
			cmdData->leftPos += len;
			cmdData->rightPos = cmdData->leftPos;
		} else
			fprintf(stderr, "NEdit: Too much binary data\n");
    }
    XtFree(outText);

    /* re-establish the timer proc (this routine) to continue processing */
    cmdData->flushTimeoutID = XtAppAddTimeOut(
    	    XtWidgetToApplicationContext(window->shell),
    	    OUTPUT_FLUSH_FREQ, flushTimeoutProc, clientData);
}

/*
** Clean up after the execution of a shell command sub-process and present
** the output/errors to the user as requested in the initial issueCommand
** call.  If "terminatedOnError" is true, don't bother trying to read the
** output, just close the i/o descriptors, free the memory, and restore the
** user interface state.
*/
static void finishCmdExecution(WindowInfo *window, int terminatedOnError)
{
    shellCmdInfo *cmdData = window->shellCmdData;
    textBuffer *buf;
    int status, failure = 0, errorReport, outTextLen, errTextLen;
    int resp, cancel = False, fromMacro = cmdData->fromMacro;
    char *outText, *errText;
	int loadAfter = (cmdData->flags & RELOAD_FILE_AFTER);

#ifndef _WIN32
    /* Cancel any pending i/o on the file descriptors */
    if (cmdData->stdoutInputID != 0) {
    	XtRemoveInput(cmdData->stdoutInputID);
		cmdData->stdoutInputID = 0;
	}
    if (cmdData->stdinInputID != 0) {
    	XtRemoveInput(cmdData->stdinInputID);
		cmdData->stdinInputID = 0;
	}
    if (cmdData->stderrInputID != 0) {
    	XtRemoveInput(cmdData->stderrInputID);
		cmdData->stderrInputID = 0;
	}
#else
	// Clean up the thread and event handles.
# if 1
	if(cmdData->stdinThreadHandle) {
		CloseHandle(cmdData->stdinThreadHandle);
		cmdData->stdinThreadHandle = 0;
	}
# endif
# if 1
	if(cmdData->stdoutThreadHandle) {
		CloseHandle(cmdData->stdoutThreadHandle);
		cmdData->stdoutThreadHandle = 0;
	}
# endif
# if 1
	if(cmdData->stderrThreadHandle) {
		CloseHandle(cmdData->stderrThreadHandle);
		cmdData->stderrThreadHandle = 0;
	}
# endif
# if 1
	if(cmdData->stdinThreadExitEventHandle) {
		CloseHandle(cmdData->stdinThreadExitEventHandle);
		cmdData->stdinThreadExitEventHandle = 0;
	}
# endif
# if 1
	if(cmdData->stdoutThreadExitEventHandle) {
		CloseHandle(cmdData->stdoutThreadExitEventHandle);
		cmdData->stdoutThreadExitEventHandle = 0;
	}
# endif
# if 1
	if(cmdData->stderrThreadExitEventHandle) {
		CloseHandle(cmdData->stderrThreadExitEventHandle);
		cmdData->stderrThreadExitEventHandle = 0;
	}
# endif
#endif

    /* Close any file descriptors remaining open */
	if(cmdData->stdoutFD != -1) {
#ifdef USE_WIN32_HANDLES
		CloseHandle(cmdData->stdoutFD);
#else
		close(cmdData->stdoutFD);
#endif
		cmdData->stdoutFD = -1;
	}
    if (cmdData->stderrFD != -1) {
#ifdef USE_WIN32_HANDLES
		CloseHandle(cmdData->stderrFD);
#else
    	close(cmdData->stderrFD);
#endif
		cmdData->stderrFD = -1;
	}
	if(cmdData->stdinFD != -1) {
#ifdef USE_WIN32_HANDLES
		CloseHandle(cmdData->stdinFD);
#else
		close(cmdData->stdinFD);
#endif
		cmdData->stdinFD = -1;
	}

    /* Free the provided input text */
    if (cmdData->input != NULL)
	XtFree(cmdData->input);
    
    /* Cancel pending timeouts */
    if (cmdData->flushTimeoutID != 0)
    	XtRemoveTimeOut(cmdData->flushTimeoutID);
    if (cmdData->bannerTimeoutID != 0)
    	XtRemoveTimeOut(cmdData->bannerTimeoutID);
    
    /* Clean up waiting-for-shell-command-to-complete mode */
    if (!cmdData->fromMacro) {
		EndWait(window->shell);
		XtSetSensitive(window->cancelShellItem, False);
		if (cmdData->bannerIsUp)
    	    ClearModeMessage(window);
    }
    
    /* If the process was killed or became inaccessable, give up */
    if (terminatedOnError) {
		freeBufList(&cmdData->outBufs);
		freeBufList(&cmdData->errBufs);
		goto cmdDone;
    }

    /* Assemble the output from the process' stderr and stdout streams into
       null terminated strings, and free the buffer lists used to collect it */
    outText = coalesceOutput(&cmdData->outBufs, &outTextLen);
    if (cmdData->flags & ERROR_DIALOGS)
    	errText = coalesceOutput(&cmdData->errBufs, &errTextLen);

    /* Wait for the child process to complete and get its return status */
#ifdef _WIN32
# if 1
	/* Wait until spawned program is done processing. */
	if(_cwait(&status, cmdData->childPid, WAIT_CHILD) == -1) {
		failure = 1;
	} else {
		failure = WIFEXITED(status) && WEXITSTATUS(status) != 0;
	}

# else	
	{
	failure = False;
	while(1)         {
		if(!GetExitCodeProcess(cmdData->childPid, &status)) {
			failure = True;
			break;
		}
		if(status != STILL_ACTIVE) {
			break;
		}
    	XtAppProcessEvent(XtWidgetToApplicationContext(window->shell), XtIMAll);
	}
	if(!failure) {
		failure = WIFEXITED(status) && WEXITSTATUS(status) != 0;
	}
	}
# endif
#else
    waitpid(cmdData->childPid, &status, 0);
	failure = WIFEXITED(status) && WEXITSTATUS(status) != 0;
#endif
    
    /* Present error and stderr-information dialogs.  If a command returned
       error output, or if the process' exit status indicated failure,
       present the information to the user. */
    if (cmdData->flags & ERROR_DIALOGS) {
	errorReport = *errText != '\0';
	if (failure && errorReport) {
    	    removeTrailingNewlines(errText);
    	    truncateString(errText, DF_MAX_MSG_LENGTH);
    	    resp = DialogF(DF_WARN, window->shell, 2, "%s",
    	    	    "Cancel", "Proceed", errText);
    	    cancel = resp == 1;
	} else if (failure) {
    	    truncateString(outText, DF_MAX_MSG_LENGTH-70);
    	    resp = DialogF(DF_WARN, window->shell, 2,
    	       "Command reported failed exit status.\nOutput from command:\n%s",
    		    "Cancel", "Proceed", outText);
    	    cancel = resp == 1;
	} else if (errorReport) {
    	    removeTrailingNewlines(errText);
    	    truncateString(errText, DF_MAX_MSG_LENGTH);
    	    resp = DialogF(DF_INF, window->shell, 2, "%s",
    	    	    "Proceed", "Cancel", errText);
    	    cancel = resp == 2;
	}
	XtFree(errText);
	if (cancel) {
	    XtFree(outText);
    	    goto cmdDone;
	}
    }
    
    /* If output is to a dialog, present the dialog.  Otherwise insert the
       (remaining) output in the text widget as requested, and move the
       insert point to the end */
    if (cmdData->flags & OUTPUT_TO_DIALOG) {
    	removeTrailingNewlines(outText);
		if (*outText != '\0') {
    		Widget widget;
    	    widget = createOutputDialog(window, outText,
    	    	loadAfter ? XmDIALOG_PRIMARY_APPLICATION_MODAL : XmDIALOG_MODELESS);

    		/* Wait for the dialog to be dismissed if loadAfter is set */
    		if(loadAfter) {
 				CheckingMode checkingModeSave;
			
				/* disable file modification checking inside the application loop
				   below. */
				checkingModeSave = window->editorInfo->checkingMode;
				window->editorInfo->checkingMode = CHECKING_MODE_DISABLED;
   				while(XtIsManaged(widget)) {
    				XtAppProcessEvent(XtWidgetToApplicationContext(window->shell), XtIMAll);
    			}
    			/* reenable modification checking */
				window->editorInfo->checkingMode = checkingModeSave;
    		}
    	}
    } else if (cmdData->flags & OUTPUT_TO_STRING) {
    	ReturnShellCommandOutput(window, outText, WEXITSTATUS(status));
    } else {
		buf = TextGetBuffer(cmdData->textW);
		if (!BufSubstituteNullChars(outText, outTextLen, buf)) {
	    	fprintf(stderr,"NEdit: Too much binary data in shell cmd output\n");
	    	outText[0] = '\0';
		}
		if (cmdData->flags & REPLACE_SELECTION) {
	    	BufReplaceSelected(buf, outText, True);
	    	TextSetCursorPos(cmdData->textW, buf->cursorPosHint);
		} else {
	    	BufReplace(buf, cmdData->leftPos, cmdData->rightPos, outText);
			if (cmdData->flags & OUTPUT_TO_NEW_WINDOW) {
    			/* If the output is to a new window then turn off the modified
    			   status so the window can be closed without being prompted to
    			   save it. Also move the cursor to the beginning of the text .*/
    			SetWindowModified(window, False);
				TextSetCursorPos(window->textArea, 0);
			} else {
	    		TextSetCursorPos(cmdData->textW, cmdData->leftPos+strlen(outText));
			}
		}
    }

    /* If the command requires the file to be reloaded afterward, reload it */
    if (loadAfter) {
		if (window->editorInfo->filenameSet && window->editorInfo->fileChanged) {
			int b;
			CheckingMode checkingModeSave;
			
			/* disable file modification checking inside the application loop
    	       in DialogF()*/
			checkingModeSave = window->editorInfo->checkingMode;
			window->editorInfo->checkingMode = CHECKING_MODE_DISABLED;
	    	b = DialogF(DF_QUES, window->shell, 2, 
	    		"Attemping to re-load file after executing command.\nThis will overwrite your changes.\nDo you want to re-load it anyway?",
    		    "Re-load", "Don't Re-load");
    		/* reenable modification checking */
			window->editorInfo->checkingMode = checkingModeSave;
    		if(b == 1) {
    			RevertToSaved(window, True);
    		} else {
    			/* If they answered no then don't prompt that the file changed */
    			char fullname[MAXPATHLEN];
    			struct stat statbuf;
    			
    			/* Get the full name of the file */
    			strcpy(fullname, window->editorInfo->path);
    			strcat(fullname, window->editorInfo->filename);
    			
    			/* get the current mtime */
    			if (stat(fullname, &statbuf) == 0) {
	   				window->editorInfo->st_mtime = statbuf.st_mtime;
    			}
    		}
    	} else {
    		RevertToSaved(window, True);
    	}
    }

    /* Command is complete, free data structure and continue macro execution */
    XtFree(outText);
cmdDone:
    XtFree((char *)cmdData);
    window->shellCmdData = NULL;
    if (fromMacro)
    	ResumeMacroExecution(window);
}

/*
** Fork a subprocess to execute a command, return file descriptors for pipes
** connected to the subprocess' stdin, stdout, and stderr streams.  cmdDir
** sets the default directory for the subprocess.  If stderrFD is passed as
** NULL, the pipe represented by stdoutFD is connected to both stdin and
** stderr.  The function value returns the pid of the new subprocess, or -1
** if an error occured.
*/
static pid_t forkCommand(Widget parent, char *command, char *cmdDir,
#ifdef USE_WIN32_HANDLES
	HANDLE *stdinFD, HANDLE *stdoutFD, HANDLE *stderrFD)
#else
	int *stdinFD, int *stdoutFD, int *stderrFD)
#endif
{
 
#ifdef _WIN32
# ifndef USE_WIN32_HANDLES
    int childStdoutFD, childStdinFD, childStderrFD, pipeFDs[2];
    pid_t childPid;
	int hStdOut, hStdIn, hStdErr;
	char cwd[_MAX_PATH];
	
	/* Create pipes to communicate with the sub process.  One end of each is
       returned to the caller, the other half is spliced to stdin, stdout
       and stderr in the child process */
#ifdef WIN32_BINARY_FILE_MODE
	if(_pipe(pipeFDs, 1, O_BINARY | O_NOINHERIT) == -1) {
#else
	if(_pipe(pipeFDs, 1, O_TEXT | O_NOINHERIT) == -1) {
#endif
		PERROR("Nedit: Internal error (opening stdout pipe)");
        return -1;
    }
    *stdoutFD = pipeFDs[0];
	/* create an inheritable copy */
    childStdoutFD = _dup(pipeFDs[1]);
	close(pipeFDs[1]);
#ifdef WIN32_BINARY_FILE_MODE
	if(_pipe(pipeFDs, 1, O_BINARY | O_NOINHERIT) == -1) {
#else
	if(_pipe(pipeFDs, 1, O_TEXT | O_NOINHERIT) == -1) {
#endif
    	PERROR("Nedit: Internal error (opening stdin pipe)");
        return -1;
    }
    *stdinFD = pipeFDs[1];
	/* create an inheritable copy */
    childStdinFD = _dup(pipeFDs[0]);
	close(pipeFDs[0]);
    if (stderrFD == NULL)
    	childStderrFD = childStdoutFD;
    else {
#ifdef WIN32_BINARY_FILE_MODE
		if(_pipe(pipeFDs, 1, O_BINARY | O_NOINHERIT) == -1) {
#else
		if(_pipe(pipeFDs, 1, O_TEXT | O_NOINHERIT) == -1) {
#endif
    	    PERROR("Nedit: Internal error (opening stderr pipe)");
            return -1;
        }
		*stderrFD = pipeFDs[0];
		/* create an inheritable copy */
		childStderrFD = _dup(pipeFDs[1]);
		close(pipeFDs[1]);
    }
    
    /* Turn off stdio buffering */
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stdin, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	// Duplicate std handles
	hStdOut = _dup(_fileno(stdout));
	hStdIn  = _dup(_fileno(stdin));
	hStdErr = _dup(_fileno(stderr));
	
	// Set the standard handles for the new process to the other end of the pipe.
	if(childStdoutFD == -1 || _dup2(childStdoutFD, _fileno(stdout)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(stdout, ...) failed:");
		return -1;
	}
	if(childStdinFD == -1 || _dup2(childStdinFD, _fileno(stdin)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(stdin, ...) failed:");
		return -1;
	}
	if(childStderrFD == -1 || _dup2(childStderrFD, _fileno(stderr)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(stderr, ...) failed:");
		return -1;
	}

	// Close the pipes handles that have been duped..
	close(childStdoutFD);
	close(childStdinFD);
    if (stderrFD != NULL) {
		close(childStderrFD);
	}

	// Remember the current directory.
	if(getcwd( cwd, _MAX_PATH ) == NULL) {
		PERROR( "_getcwd error" );
		return -1;
	}

	/* change the current working directory to the directory of the current
	   file. */ 
	if(cmdDir[0] != 0) {
		if(chdir(cmdDir) == -1) {
			PERROR("chdir to directory of current file failed");
			return -1;
		}
	}
	
	_flushall();
	
	// Spawn process
	childPid = _spawnlp(_P_NOWAIT, GetPrefShell(), GetPrefShell(), "-c", command, NULL);
	
	if(childPid == -1) {
		childPid = _spawnlp(_P_NOWAIT, "cmd", "cmd", "/c", command, NULL);
	}

	// Restore the current directory.
	if(chdir(cwd) == -1) {
		PERROR("restore of the current directory failed");
		return -1;
	}
	
	if(hStdOut != -1 && _dup2(hStdOut, _fileno(stdout)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(hStdOut) failed:");
		return -1;
	}
	if(hStdIn != -1 && _dup2(hStdIn, _fileno(stdin)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(hStdIn) failed:");
		return -1;
	}
	if(hStdErr != -1 && _dup2(hStdErr, _fileno(stderr)) == -1) {
   	    PERROR("Nedit: Internal error _dup2(hStdErr) failed:");
		return -1;
	}

	close(hStdOut);
	close(hStdIn);
	close(hStdErr);

# else
	HANDLE childStdoutFD, childStdinFD, childStderrFD;
	HANDLE childStdoutFDDup, childStdinFDDup, childStderrFDDup;
    pid_t childPid;
 	HANDLE hStdOut, hStdIn, hStdErr;
	char cwd[_MAX_PATH];
	SECURITY_ATTRIBUTES saAttr;
	
	// Set the bInheritHandle flag so pipe handles are inherited.
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.bInheritHandle = TRUE;
	saAttr.lpSecurityDescriptor = NULL;
	
	// The steps for redirecting child process's STDOUT:
	//     1. Save current STDOUT, to be restored later.
	//     2. Create anonymous pipe to be STDOUT for child process.
	//     3. Set STDOUT of the parent process to be write handle to
	//        the pipe, so it is inherited by the child process.
	//     4. Create a noninheritable duplicate of the read handle and
	//        close the inheritable read handle.
	
	// Save the current standard handles.
	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	hStdIn  = GetStdHandle(STD_INPUT_HANDLE);
	hStdErr = GetStdHandle(STD_ERROR_HANDLE);
	
	// Create a pipe for the child process's STDOUT.
	if (! CreatePipe((void**)stdoutFD, &childStdoutFD, &saAttr, IO_BUF_SIZE)) {
		PERROR("Nedit: Internal error (opening stdout pipe)");
		*stdoutFD = -1;
        return -1;
    }
	if (! CreatePipe((void**)stdinFD, &childStdinFD, &saAttr, IO_BUF_SIZE)) {
		PERROR("Nedit: Internal error (opening stdin pipe)");
		*stdinFD = -1;
        return -1;
    }
    if (stderrFD == NULL)
    	childStderrFD = childStdoutFD;
    else {
		if (! CreatePipe((void**)stderrFD, &childStderrFD, &saAttr, IO_BUF_SIZE)) {
			PERROR("Nedit: Internal error (opening stderr pipe)");
			*stderrFD = -1;
			return -1;
	    }
	}
	
	// Set the standard handles for the new process to the other end of the pipe.
	if(!SetStdHandle(STD_OUTPUT_HANDLE, childStdoutFD)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_OUTPUT_HANDLE, ...) failed:");
		return -1;
	}
	if(!SetStdHandle(STD_INPUT_HANDLE, childStdinFD)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_INPUT_HANDLE, ...) failed:");
		return -1;
	}
	if(!SetStdHandle(STD_ERROR_HANDLE, childStderrFD)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_ERROR_HANDLE, ...) failed:");
		return -1;
	}

	// Create noninheritable read handle and close the inheritable read 
	// handle.
	if(!DuplicateHandle(
			GetCurrentProcess(), childStdoutFD,
			GetCurrentProcess(), &childStdoutFDDup,
			0, FALSE, DUPLICATE_SAME_ACCESS
	    )
	) {
   	    PERROR("Nedit: Internal error DuplicateHandle for stdout failed:");
		return -1;
	}
	if(!DuplicateHandle(
			GetCurrentProcess(), childStdinFD,
			GetCurrentProcess(), &childStdinFDDup,
			0, FALSE, DUPLICATE_SAME_ACCESS
	    )
	) {
   	    PERROR("Nedit: Internal error DuplicateHandle for stdin failed:");
		return -1;
	}
	if(!DuplicateHandle(
			GetCurrentProcess(), childStderrFD,
			GetCurrentProcess(), &childStderrFDDup,
			0, FALSE, DUPLICATE_SAME_ACCESS
	    )
	) {
   	    PERROR("Nedit: Internal error DuplicateHandle for stderr failed:");
		return -1;
	}

	CloseHandle(childStdoutFD);
	CloseHandle(childStdinFD);
    if (stderrFD != NULL)
		CloseHandle(childStderrFD);

	// Remember the current directory.
	if(getcwd( cwd, _MAX_PATH ) == NULL) {
		PERROR( "_getcwd error" );
		return -1;
	}

	/* change the current working directory to the directory of the current
	   file. */ 
	if(cmdDir[0] != 0) {
		if(chdir(cmdDir) == -1) {
			PERROR("chdir to directory of current file failed");
			return -1;
		}
	}
	
	_flushall();
	
	// Spawn process
	childPid = _spawnlp(_P_NOWAIT, "cmd", "cmd", "/c", command, NULL);

	// Restore the current directory.
	if(chdir(cwd) == -1) {
		PERROR("restore of the current directory failed");
		return -1;
	}
	
	if(hStdOut != INVALID_HANDLE_VALUE && !SetStdHandle(STD_OUTPUT_HANDLE, hStdOut)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_OUTPUT_HANDLE, hStdOut) failed:");
		return -1;
	}
	if(hStdIn != INVALID_HANDLE_VALUE && !SetStdHandle(STD_INPUT_HANDLE, hStdIn)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_INPUT_HANDLE, hStdIn) failed:");
		return -1;
	}
	if(hStdErr != INVALID_HANDLE_VALUE && !SetStdHandle(STD_ERROR_HANDLE, hStdErr)) {
   	    PERROR("Nedit: Internal error SetStdHandle(STD_ERROR_HANDLE, hStdErr) failed:");
		return -1;
	}
# endif
#else
    int childStdoutFD, childStdinFD, childStderrFD, pipeFDs[2];
    int dupFD;
    pid_t childPid;
    
    /* Create pipes to communicate with the sub process.  One end of each is
       returned to the caller, the other half is spliced to stdin, stdout
       and stderr in the child process */
    if (pipe(pipeFDs) != 0) {
		PERROR("Nedit: Internal error (opening stdout pipe)");
        return -1;
    }
    *stdoutFD = pipeFDs[0];
    childStdoutFD = pipeFDs[1];
    if (pipe(pipeFDs) != 0) {
    	PERROR("Nedit: Internal error (opening stdin pipe)");
        return -1;
    }
    *stdinFD = pipeFDs[1];
    childStdinFD = pipeFDs[0];
    if (stderrFD == NULL)
    	childStderrFD = childStdoutFD;
    else {
	if (pipe(pipeFDs) != 0) {
    	    PERROR("Nedit: Internal error (opening stderr pipe)");
            return -1;
        }
	*stderrFD = pipeFDs[0];
	childStderrFD = pipeFDs[1];
    }
    
    /* Fork the process */
    childPid = fork();
    
    /*
    ** Child process context (fork returned 0), clean up the
    ** child ends of the pipes and execute the command
    */
    if (0 == childPid) {

	/* Close the connection to the X display in the child. */
	close(ConnectionNumber(TheDisplay));
	
	/* close the parent end of the pipes in the child process   */
	close(*stdinFD);
	close(*stdoutFD);
	if (stderrFD != NULL)
	    close(*stderrFD);

	/* close current stdin, stdout, and stderr file descriptors before
	   substituting pipes */
	close(fileno(stdin));
	close(fileno(stdout));
	close(fileno(stderr));

	/* duplicate the child ends of the pipes to have the same numbers
	   as stdout & stderr, so it can substitute for stdout & stderr */
 	dupFD = dup2(childStdinFD, fileno(stdin));
	if (dupFD == -1) {
	    PERROR("dup of stdin failed");
		return 1;
	}
 	dupFD = dup2(childStdoutFD, fileno(stdout));
	if (dupFD == -1) {
	    PERROR("dup of stdout failed");
		return 1;
	}
 	dupFD = dup2(childStderrFD, fileno(stderr));
	if (dupFD == -1) {
	    PERROR("dup of stderr failed");
		return 1;
	}
	
	/* make this process the leader of a new process group, so the sub
	   processes can be killed, if necessary, with a killpg call */
	setsid();
	
	/* change the current working directory to the directory of the current
	   file. */ 
	if(cmdDir[0] != 0) {
		if(chdir(cmdDir) == -1) {
			PERROR("chdir to directory of current file failed");
			return 1;
		}
	}
	
	/* execute the command using the shell specified by preferences */
	execl(GetPrefShell(), GetPrefShell(), "-c", command, 0);

	fprintf(stderr, 
		"Error starting preferred shell: %s\nUsing /bin/sh as a last resort...\n",
		GetPrefShell());
	
	/* if that failed try /bin/sh as a last resort. */
	execl("/bin/sh", "/bin/sh", "-c", command, 0);

	/* if we reach here, execl failed */
	fprintf(stderr, "Error starting shell: %s\n", "/bin/sh");
	exit(1);
    }
    
    /* close the child ends of the pipes */
    close(childStdinFD);
    close(childStdoutFD);
    if (stderrFD != NULL)
    	close(childStderrFD);

#endif
    /* Parent process context, check if fork succeeded */
    if (childPid == -1)
    	DialogF(DF_ERR, parent, 1,
		"Error starting shell command process\n(fork/spawn failed)",
		"Dismiss");

    return childPid;
}    

/*
** Add a buffer full of output to a buffer list
*/
static void addOutput(buffer **bufList, buffer *buf)
{
    buf->next = *bufList;
    *bufList = buf;
}

/*
** coalesce the contents of a list of buffers into a contiguous memory block,
** freeing the memory occupied by the buffer list.  Returns the memory block
** as the function result, and its length as parameter "length".
*/
static char *coalesceOutput(buffer **bufList, int *outLength)
{
    buffer *buf, *rBufList = NULL;
    char *outBuf, *outPtr, *p;
    int i, length = 0;
    
    /* find the total length of data read */
    for (buf=*bufList; buf!=NULL; buf=buf->next)
    	length += buf->length;
    
    /* allocate contiguous memory for returning data */
    outBuf = XtMalloc(length+1);
    
    /* reverse the buffer list */
    while (*bufList != NULL) {
    	buf = *bufList;
    	*bufList = buf->next;
    	buf->next = rBufList;
    	rBufList = buf;
    }
    
    /* copy the buffers into the output buffer */
    outPtr = outBuf;
    for (buf=rBufList; buf!=NULL; buf=buf->next) {
    	p = buf->contents;
    	for (i=0; i<buf->length; i++)
    	    *outPtr++ = *p++;
    }
    
    /* terminate with a null */
    *outPtr = '\0';

    /* free the buffer list */
    freeBufList(&rBufList);
    
    *outLength = outPtr - outBuf;
    return outBuf;
}

static void freeBufList(buffer **bufList)
{
    buffer *buf;
    
    while (*bufList != NULL) {
    	buf = *bufList;
    	*bufList = buf->next;
    	XtFree((char *)buf);
    }
}

/*
** Remove trailing newlines from a string by substituting nulls
*/
static void removeTrailingNewlines(char *string)
{
    char *endPtr = &string[strlen(string)-1];
    
    while (endPtr >= string && *endPtr == '\n')
    	*endPtr-- = '\0';
}

/*
** Create a dialog for the output of a shell command.  The dialog lives until
** the user presses the Dismiss button, and is then destroyed
*/
static Widget createOutputDialog(WindowInfo *window, char *text, 
	unsigned char dialogStyle)
{
    Arg al[50];
    int ac, rows, cols, hasScrollBar;
    Widget form, textW, button;
    XmString st1;

    /* measure the width and height of the text to determine size for dialog */
    measureText(text, MAX_OUT_DIALOG_COLS, &rows, &cols);
    if (rows > MAX_OUT_DIALOG_ROWS) {
    	rows = MAX_OUT_DIALOG_ROWS;
    	hasScrollBar = True;
    } else
    	hasScrollBar = False;
    if (cols > MAX_OUT_DIALOG_COLS)
    	cols = MAX_OUT_DIALOG_COLS;
    if (cols == 0)
    	cols = 1;
    
    ac = 0;
    XtSetArg(al[ac], XmNdialogStyle, dialogStyle); ac++;
    form = XmCreateFormDialog(window->shell, "shellOutForm", al, ac);

    ac = 0;
    XtSetArg(al[ac], XmNlabelString, st1=MKSTRING("Dismiss")); ac++;
    XtSetArg(al[ac], XmNhighlightThickness, 0);  ac++;
    XtSetArg(al[ac], XmNbottomAttachment, XmATTACH_FORM);  ac++;
    XtSetArg(al[ac], XmNtopAttachment, XmATTACH_NONE);  ac++;
    button = XmCreatePushButtonGadget(form, "dismiss", al, ac);
    XtManageChild(button);
    XtVaSetValues(form, XmNdefaultButton, button, 0);
    XmStringFree(st1);
    XtAddCallback(button, XmNactivateCallback, destroyOutDialogCB,
    	    XtParent(form));
    
	{
	Widget sw;
    int emTabDist, wrapMargin;
    char *delimiters;
    /* Create a text widget inside of a scrolled window widget */
    sw = XtVaCreateManagedWidget("scrolledW", xmScrolledWindowWidgetClass, form,
    	    XmNspacing, 0, 
#if 0
    	    XmNpaneMaximum, SHRT_MAX,
    	    XmNpaneMinimum, PANE_MIN_HEIGHT, 
#endif
    	    XmNhighlightThickness, 0, 
    	    XmNtopAttachment, XmATTACH_FORM,
    	    XmNleftAttachment, XmATTACH_FORM,
    	    XmNbottomAttachment, XmATTACH_WIDGET,
    	    XmNrightAttachment, XmATTACH_FORM,
    	    XmNbottomWidget, button,
    	    0); 
    XtVaGetValues(window->textArea,
    		textNemulateTabs, &emTabDist,
    		textNwrapMargin, &wrapMargin, 
    		textNwordDelimiters, &delimiters, 
    		0);
    textW = XtVaCreateManagedWidget("text", textWidgetClass, sw,
    	    textNrows, rows, 
    	    textNcolumns, cols,
    	    textNfont, GetDefaultFontStruct(window->fontList),
    		textNemulateTabs, emTabDist,
    		textNwrapMargin, wrapMargin, 
    		textNwordDelimiters, delimiters, 
    	    textNautoIndent, window->editorInfo->indentStyle == AUTO_INDENT,
    	    textNsmartIndent, window->editorInfo->indentStyle == SMART_INDENT,
    	    textNautoWrap, window->editorInfo->wrapMode == NEWLINE_WRAP,
    	    textNcontinuousWrap, window->editorInfo->wrapMode == CONTINUOUS_WRAP,
    	    textNoverstrike, window->editorInfo->overstrike,
    	    0);
    TextSetString(textW, text);
    }
    
    XtVaSetValues(XtParent(form), XmNtitle, "Nedit - Output from Command", 0);
    ManageDialogCenteredOnPointer(form);
    return form;
}

/*
** Dispose of the command output dialog when user presses Dismiss button
*/
static void destroyOutDialogCB(Widget w, XtPointer callback, XtPointer closure)
{
    XtDestroyWidget((Widget)callback);
}

/*
** Measure the width and height of a string of text.  Assumes 8 character
** tabs.  wrapWidth specifies a number of columns at which text wraps.
*/
static void measureText(char *text, int wrapWidth, int *rows, int *cols)
{
    int maxCols = 0, line = 0, col = 0;
    char *c;
    
    for (c=text; *c!='\0'; c++) {
    	if (*c=='\n' || col > wrapWidth) {
    	    line++;
    	    col = 0;
    	} else {
    	    if (*c == '\t')
    		col += 8 - (col % 8);
    	    else
    		col++;
    	    if (col > maxCols)
    	    	maxCols = col;
    	}
    }
    /* If there is no newline before the end of the string then add
    ** another line for the end of the string.
    */
    if(c[-1] != '\n') line++;
    
    *rows = line;
    *cols = maxCols;
}

/*
** Truncate a string to a maximum of length characters.  If it shortens the
** string, it appends "..." to show that it has been shortened. It assumes
** that the string that it is passed is writeable.
*/
static void truncateString(char *string, int length)
{
    if (strlen(string) > length)
	memcpy(&string[length-3], "...", 4);
}

/*
** Substitute the string subsStr in inStr wherever % appears, storing the
** result in outStr.  Returns False if the resulting string would be
** longer than outLen
*/
/* .b: argument change */
static int substitutePercent(WindowInfo *window, char *outStr, char *inStr,
        int outLen, char *selection)
/* .e */
{
    char *inChar, *outChar, *c;

    inChar = inStr;
    outChar = outStr;
    while (*inChar != '\0') {
    	/* pass an escaped % to the output */
    	if (*inChar == '\\' && inChar[1] == '%') {
    		inChar++;
    	    *outChar++ = *inChar++;
    	}
    	if (*inChar == '%') {
			/* .b: Replace "%t" and parse "%:message:default_value%" syntax */
			if (*(inChar+1) == '%') { /* %% is % */
    	    	inChar += 2;
    	    	*outChar++ = '%';
    	    } else if (*(inChar+1) == 't') { /* %t is tabDist */
    	        sprintf(outChar, "%d", window->editorInfo->buffer->tabDist);
    	        inChar += 2;
                outChar += strlen(outChar);
    	    } else if (*(inChar+1) == 's') { /* %s is the selection contents */
				if(selection) {
    				strcpy(outChar, selection);
	                outChar += strlen(outChar);
				}
    	        inChar += 2;
    	    } else if (*(inChar+1) == 'd') { /* %d is the directory */
   				strcpy(outChar, window->editorInfo->path);
                outChar += strlen(outChar);
    	        inChar += 2;
    	    } else if (*(inChar+1) == 'f') { /* %f is the basename */
   				strcpy(outChar, window->editorInfo->filename);
                outChar += strlen(outChar);
    	        inChar += 2;
    	    } else if (*(inChar+1) == ':') { /* Ask for arguments */

    		/* Saves outChar position */
    		c = outChar;

                /* Creates dialog_message */
    		for (inChar += 2; *inChar != ':'; inChar++, outChar++) {
    		    if (*inChar == '\0') {
    		        DialogF(DF_ERR, window->shell, 1,
	                       "Syntax error: Separator ':' not found", "OK");
	                return False;
	            } else if (*inChar == '\\') {
	                inChar++;
	            }
	            *outChar = *inChar;
   	    	}
   	    	*outChar++ = ':';
		*outChar++ = '\0';
		
		/* Creates default_value */
    		for (inChar++; *inChar != '%'; inChar++, outChar++) {
    		    if (*inChar == '\0') {
    		        DialogF(DF_ERR, window->shell, 1,
	                       "Syntax error: Final '%%' not found", "OK");
	                return False;
	            } else if (*inChar == '\\') {
	                inChar++;
	            }
	            *outChar = *inChar;
   	    	}
		inChar++;
		*outChar = '\0';

		/* Display request dialog */
		if (DialogF(DF_PROMPT2, window->shell, 2,c,
		            c+strlen(c)+1, c, "OK", "Cancel") == 2)
        	    return False;

    		outChar=outStr+strlen(outStr);
    	    } else {
				/* Substitute the current file name for % in the shell command */
				strcpy(outChar, window->editorInfo->path);
				strcat(outChar, window->editorInfo->filename);
    		inChar++;
				if(*inChar == 'p') inChar++; /* %p is the pathname */
	   			outChar=outStr+strlen(outStr);
    	    }
    	} else
    	    *outChar++ = *inChar++;

    	if (outChar - outStr >= outLen) {
            DialogF(DF_ERR, window->shell, 1,
	"Shell command is too long due to\nfilename substitutions with '%%'",
	            "OK");
    	    return False;
        }
    }
    *outChar = '\0';
    return True;
}
